"""
    GraphQL (Schema + Operations)
"""

import json
import os.path

import click

from ... import Schema, Settings

settings = Settings()


def to_camel_case(text):
    """Converts to Camel-Case"""
    init, *temp = text.split("_")
    return "".join([init.lower(), *map(str.title, temp)])


def read_files(root_path):
    """Read All .graphql Files"""
    outputs = []
    if os.path.isdir(root_path):
        for item in root_path.glob("**/*.graphql"):
            with open(item, "r", encoding="utf-8") as file:
                code = file.read()
                outputs.append(code)
    return outputs


def get_operations(base_dir):
    """Get App GraphQL Operations"""
    # Operations <Files>
    operations_location = base_dir / "operations"
    operations_desktop = operations_location / "desktop"
    operations_mobile = operations_location / "mobile"

    # Return Values
    outputs = dict(desktop=[], mobile=[])
    outputs["desktop"].extend(read_files(operations_desktop))
    outputs["mobile"].extend(read_files(operations_mobile))

    # Return
    return outputs


def shell_print(text: str, color: str = "green"):
    """Customized Click Echo"""
    return click.secho(f"{ text }", fg=color, bold=True)


@click.command(name="schema")
def graphql():
    """Build GraphQL (Schema + Operations)"""

    # Get Path(s)
    output_dir = settings.base_dir
    for path in settings.base.generates.split("/"):
        output_dir = output_dir / path

    # Create Path(s)
    output_dir.mkdir(parents=True, exist_ok=True)

    # Output Paths <Files>
    desktop_operations = output_dir / "desktop.graphql"
    mobile_operations = output_dir / "mobile.graphql"
    schema_graphql = output_dir / "schema.graphql"
    schema_json = output_dir / "operations.json"

    # Outputs
    shell_print("* Collecting All Operations From...", color="magenta")
    shell_print("\t * <app_name>/operations/desktop...", color="magenta")
    shell_print("\t * <app_name>/operations/mobile... \n", color="magenta")
    outputs = dict(desktop=[], mobile=[])
    for path in settings.apps.paths:
        code = get_operations(path)
        outputs["desktop"].extend(code["desktop"])
        outputs["mobile"].extend(code["mobile"])

    # Finally
    desktop = "\n\n".join(outputs["desktop"])
    mobile = "\n\n".join(outputs["mobile"])

    shell_print("* Building Schema & Operations ...")
    shell_print(f"\t * [schema]: { schema_graphql }")
    shell_print(f"\t * [desktop]: { desktop_operations }")
    shell_print(f"\t * [mobile]: { mobile_operations }")
    shell_print(f"\t * [operations]: { schema_json }")

    # Desktop (Operations)
    with open(desktop_operations, "w", encoding="utf-8") as file:
        file.write(desktop)

    # Mobile (Operations)
    with open(mobile_operations, "w", encoding="utf-8") as file:
        file.write(mobile)

    # GraphQL (Schema)
    schema = Schema(
        query=settings.apps.schema.Query,
        mutation=settings.apps.schema.Mutation,
    )
    with open(schema_graphql, "w", encoding="utf-8") as file:
        file.write(str(schema))

    # Names (Operations)
    query_names = settings.apps.operations.query
    mutation_names = settings.apps.operations.mutation
    query_names_camel = set()
    mutation_names_camel = set()

    for x in query_names:
        query_names_camel.add(to_camel_case(x))
    for x in mutation_names:
        mutation_names_camel.add(to_camel_case(x))

    # Pretty (Operations)
    pretty = lambda items: sorted(list(set(items)))
    operations_names = dict(
        python=dict(query=pretty(query_names), mutation=pretty(mutation_names)),
        graphql=dict(
            query=pretty(query_names_camel), mutation=pretty(mutation_names_camel)
        ),
    )

    # Output (Operations)
    with open(schema_json, "w", encoding="utf-8") as file:
        file.write(json.dumps(operations_names, indent=4))



"""
    Fastberry Main (Core-Methods)
"""

from starlette.middleware.base import BaseHTTPMiddleware as BaseMiddleware

from . import config
from .config import Settings
from .config.extras.crud import CRUD
from .databases import Database
from .graphql import BaseType
from .schema import Schema

# BaseType, CRUD, Database, Schema, Settings



"""
    Project Core Definitions (aka: Folders & Files)
"""

import json
import types

import yaml

from .settings_base import Settings

FILE_YAML = "settings.yaml"
FILE_MODE = "config/mode.json"
FILE_DOCS = "config/docs.md"
LOAD_ENVS = lambda mode: f"config/env/{ mode }.env"
FILE_ENVS = "config/env/development.env"

SETTING_YAML_QUERYING = {"items_per_page": 50}
SETTING_YAML = {
    "version": "0.0.1",
    "app_name": "FastAPI",
    "admin_email": None,
    "installed_apps": [],
    "development_apps": [],
    "querying": {},
    "allowed_hosts": [],
    "permissions": [],
    "middleware": [],
    "generates": "",
}


def load_yaml():
    """Load <Yaml> File"""
    return_value = None
    with open(FILE_YAML, "r", encoding="utf-8") as stream:
        try:
            settings_yaml = yaml.safe_load(stream)
            settings_dir = {name.lower(): conf for name, conf in settings_yaml.items()}
            # Other Dicts
            SETTING_YAML_QUERYING.update(settings_dir.get("querying", {}))
            settings_dir.update({"querying": SETTING_YAML_QUERYING})
            SETTING_YAML.update(settings_dir)
            return_value = types.SimpleNamespace(**SETTING_YAML)
        except yaml.YAMLError as error:
            print(error)
    return return_value


def load_mode():
    """Load <Yaml> File"""
    return_value = {}
    with open(FILE_MODE, "r", encoding="utf-8") as file:
        try:
            data = file.read()
            if data:
                return_value = json.loads(data)
        except yaml.YAMLError as error:
            print(error)
    return return_value


def load_env(mode: str = None):
    """Load <Dot.Env> File"""
    active_env = LOAD_ENVS(mode) if mode else FILE_ENVS
    return Settings(_env_file=active_env)


def load_docs():
    """Load <Docs> File"""
    DOCS = ""
    with open(FILE_DOCS, "r", encoding="utf-8") as file:
        DOCS = file.read()
    return DOCS



"""
    BaseSettings created with Pydantic
"""

from pydantic import BaseSettings


class Settings(BaseSettings):
    """Core Settings"""

    secret_key: str
    debug: bool = False
    mongo: str | None
    sql: str | None

    class Config:
        """Default Environmental File"""

        env_file = ".env"



"""
    Create A New App
"""

import os.path
import pathlib

import click

from ... import Settings
from .shell import shell_print, unzip

TEMPLATES_DIR = pathlib.Path(__file__).parents[0] / "templates"

settings = Settings()


@click.command()
@click.argument(
    "app_name",
    type=str,
    nargs=1,
)
def start_app(app_name):
    """Creates a Fastberry App Directory."""

    # Get Path(s)
    apps_dir = settings.base_dir / "apps"
    the_dir = apps_dir / app_name

    # Create Path(s)
    apps_dir.mkdir(parents=True, exist_ok=True)

    # Check Path(s)
    if os.path.isdir(the_dir):
        shell_print(f"""* App: "{ app_name }" Already Exists!""", color="red")
    else:
        shell_print(
            f"""* Starting App: "{ app_name }" ...""",
        )
        unzip(TEMPLATES_DIR / "base_app.zip", the_dir)



"""
    Auto-Importer
"""

import importlib
import pkgutil

from ..utils.objects import get_attr


def iter_namespace(ns_pkg):
    """https://packaging.python.org/en/latest/guides/creating-and-discovering-plugins/"""
    return pkgutil.iter_modules(ns_pkg.__path__, ns_pkg.__name__ + ".")


def import_module(single_app: str):
    """Import Single-Module"""
    try:
        module = importlib.import_module(single_app)
    except:
        module = None
    return module


def import_modules(all_apps: list):
    """https://packaging.python.org/en/latest/guides/creating-and-discovering-plugins/"""
    installed_apps = {}
    for app in all_apps:
        module = import_module(app)
        if module:
            discovered_plugins = {
                name: importlib.import_module(name)
                for finder, name, ispkg in iter_namespace(module)
            }
            installed_apps.update(discovered_plugins)
    return installed_apps


def search_method(dotted_path: str):
    """Search for Method in <Module>"""
    parts = dotted_path.split(".")
    root = import_module(parts[0])
    module = root
    for part in parts[1:]:
        module = get_attr(module, part)
    return module
